
/* Portions Copyright (c) 1995,1996,1997 by Trusted Information Systems, Inc.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND TRUSTED INFORMATION SYSTEMS DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL TRUSTED INFORMATION 
 * SYSTEMS BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 *
 * Trusted Information Systems, Inc. has received approval from the
 * United States Government for export and reexport of TIS/DNSSEC
 * software from the United States of America under the provisions of
 * the Export Administration Regulations (EAR) General Software Note
 * (GSN) license exception for mass market software.  Under the
 * provisions of this license, this software may be exported or
 * reexported to all destinations except for the embargoed countries of
 * Cuba, Iran, Iraq, Libya, North Korea, Sudan and Syria.  Any export
 * or reexport of TIS/DNSSEC software to the embargoed countries
 * requires additional, specific licensing approval from the United
 * States Government. 
 */




static char *rcsid = "$Header: /m/a/projects/iip/iip/sec_bind494/dnssec_lib/RCS/dnssec_api.c,v 1.24 1997/01/23 15:49:10 ogud Exp $";
/*
 * This file contains the interface between the DNSSEC API to the crypto 
 * API. 
 * This is the only file that needs to be changed if the crypto system is 
 * changed.
 * exported functions are 
 * int  dnssec_check()     Function to check if certain alg is suppored
 * int  dnssec_sign():     Incremental signing routine takes flags to 
 *                         select which steps to perform 
 * int  dnssec_verify()    Incremental verify routine
 * int  dnssec_genkey()    Function to generate new KEY for dns
 * KEY *dnssec_getkey()    Function to retrive private KEY from storage
 *                         private keys are stored internaly and are only 
 *                         decoded once.
 * void dnssec_putkey()    Function to write out a private key 
 * 
 * KEY *dnssec_pubkey()    Function to convert KEY RR into a KEY structure
 * KEY *dnssec_readpubkey()  Funcation that reads in public key 
 * int dnssec_writepubkey()  Function to write out public key 
 * int dnssec_encpubkey()   Function to return a public key in DNS format
 *                             base64 encoded
 *
 * void dnssec_freekey()   Releases all memory pointed to by key structure
 */


#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <memory.h>
#include <sys/param.h> 
#include <netinet/in.h>  
#include <arpa/nameser.h>
#ifdef USE_OPTIONS_H
# include "../conf/options.h"
# include "../conf/portability.h"
#else
# ifndef PATH_MAX
#  ifdef POSIX_PATH_MAX
#   define PATH_MAX POSIX_PATH_MAX
#  else
#   define PATH_MAX 255 /* this is the value of POSIX_PATH_MAX */
#  endif
# endif 
#endif
#ifdef DNS_SECURITY
# include "ns_defs.h"
# include "db_defs.h"
#endif
#include "dnssec.h" 
#include "dns_crypto.h"

#ifndef DUMP
#define DUMP_COND if(mode & SIG_MODE_AXFR)
# ifdef LONG_TEST
#  define DUMP(data, mode, len, msg) DUMP_COND {   char x[1000];\
	 			int n=encode64( data, x, len);\
				printf("%s: %x %d %s\n",msg, mode, n, x);}
# else
#  ifdef SHORT_TEST
#   define DUMP(a,mode, len, msg) DUMP_COND printf("%s,%x %d\n",msg, mode, len)
#  else 
#   define DUMP(a,b,c, d)
#  endif /* SHORT_TEST */
# endif  /* LONG_TEST */
#endif  /* DUMP */

/* Local non exported functions  */
static int    dnssec_decrypt    __P((u_char *, u_char *, int));
static int    dnssec_encrypt    __P((u_char *, u_char *, int));
static void   dnssec_putkey     __P((KEY *));

/* Local non exported functions that do not fit the naming schema */
static int    ReadPrivateKey    __P((char *, KEY *));
static int    WritePublicKey    __P((KEY *, int));
static void   strip_trailing_dots __P(( char *));

/* 
 * dnssec_check()  function to check if crypto system for specified 
 *   algorithm is present. 
 * Return 1 if it is 
 *        0 if not present or not supported 
 */
int 
dnssec_check( alg)
int alg; 
{
  switch( alg) {
    case KEY_RSA:
#ifdef RSAREF
       return 1; 
#endif /* RSAREF */
       break;
 /* add other supported alg here */
    default:
       break;
  }
  return 0; 
}

/*
 * Function to compare two keys to see if they are equal 
 * Returns 0 if equal 
 *         non zero if different 
 */ 
int
dnssec_keycmp(key1, key2)
KEY *key1, *key2; 
{
  if( key1->alg != key2->alg)
    return 1;
  if( key1->key_size != key2->key_size) 
    return 2;
  if( key1->foot != key2->foot) 
    return 3;
  switch( key1->alg) {
    case KEY_RSA:
#ifdef RSAREF 
    return rsa_cmp( (RSA_Key *) key1->KEY_struct, (RSA_Key *)key2->KEY_struct);
#endif
   default:
      return memcmp( key1->KEY_struct, key2->KEY_struct, key2->key_size);
    }
}

/*
 * Incremental signing function 
 * If len and data are non NULL the contents of data is added to the hash
 * even if the flags are INIT or FINAL
 * Parameters: 
 *     mode  what operations to perform this time. 
 *     data  the data to pass through the hash function
 *     len   the length of the data 
 *     in_key the key to use 
 *     signature the location to write the signature to 
 * Return codes
 *    0 success INIT or Update operation 
 *    positive value success FINAL (sign)  operation
 *    negative value failure 
 */ 

int 
dnssec_sign(mode, data, len,  in_key, signature)
int mode, len;
u_char *data;
KEY *in_key;
u_char *signature;
{
  int sign_len=0;
  int axfr =  mode & SIG_MODE_AXFR ? 1 : 0;
  DUMP(data, mode, len, "dnssec_sign()");
  switch (in_key->alg) {
    case KEY_RSA: 
#ifdef RSAREF
    {
      RSA_Key *key;
      int i,j; 
      key = (RSA_Key *) in_key->KEY_struct;
      sign_len = rsa_sign( mode, &key->context[axfr], data, len, 
			  key->Private_Key,   signature);
      if( sign_len > 0 && mode & SIG_MODE_FINAL) { 
	for( i=0;signature[i]== 0x0 && i < sign_len ; i++);
	if( i>0) { 
	  printf("dnssec_sign: stripping %d null Bytes of SIGNATURE\n",i);
	  for( j=0; i < sign_len; j++,i++)  	  /* do a safe memcopy */
	    signature[j] = signature[i];
	  sign_len = j;
	}
      }
      return sign_len;
    }
#endif /*RSAREF */
    default: 
         break;
  }
  return UNKNOWN_KEYALG;
}
/*
 * Incremental verification routine
 * This routine will add data to the hash on every call if any data is passed
 * in even when the mode is INIT or FINAL 
 * Parameters: 
 *     mode  what operations to perform this time. 
 *     data  the data to pass through the hash function
 *     len   the length of the data 
 *     in_key the key to use 
 *     signature the location to write the signature to 
 *     sig_len  lenght of the signature in bytes.
 * Return codes
 *    0 success 
 *    any other value failure 
 */
int 
dnssec_verify(mode, data, len,  in_key, signature, sig_len)
int mode,len, sig_len;
u_char *data;
KEY *in_key;
u_char *signature;
{
  int axfr =  mode & SIG_MODE_AXFR ? 1: 0;
  DUMP(data, mode, len, "dnssec_verify()");
  switch (in_key->alg) {
  case KEY_RSA: 
#ifdef RSAREF
    {
      RSA_Key *key = (RSA_Key *) in_key->KEY_struct;
      return rsa_verify( mode, &key->context[axfr], data, len, 
			  key->Public_Key, signature, sig_len);
    }
#endif /* RSAREF */
    default:
      break;
  }
  return UNKNOWN_KEYALG;
}

/*
 * Key Functions
 * Function to get private key 
 * Parameters 
 *      keyname  the name of the KEY
 *      force_read if true forces that the key be read from disk. 
 * 
 * Return
 *      NULL     if there is no key found in the current directory 
 *               or this key has not been loaded before 
 *     !NULL     KEY structure filled out 
 */
static struct key_chain {
    KEY *key;
    struct key_chain *next; 
} *chain_head=NULL;
KEY * 
dnssec_getkey( keyname, force_read ) 
char * keyname;
int force_read;
{
  KEY *dg_key=NULL, *pubkey=NULL; 
  int t;
  struct key_chain *next_key=NULL;
  /* look for key among keys that have already been read in,  I need to
     purge the chain once in a while, each key needs an expiration time
  */ 
  if(keyname == NULL) 
    return NULL; 
  strip_trailing_dots(keyname);
  t= strlen(keyname);
  if( force_read != DNSSEC_FORCE_READ) {
    for( next_key = chain_head; next_key && next_key->key; 
	next_key=next_key->next)
      if(strcasecmp( keyname, next_key->key->key_name)==0 && 
	 t == strlen(next_key->key->key_name)) break;
    
    if( next_key)  /* found the key */ 
      return next_key->key;
  }
  /* before I read in the key check if it is allowed to sign */ 
  if( (pubkey=dnssec_readpubkey(keyname)) == NULL)
    return NULL;

/* allowed to sign */
  if(((pubkey->flags & (SIGN_F_SIGNATORY | SIGN_F_ZONE)) == 0) ||
      ((pubkey->flags & SIGN_F_NOAUTHENT) == SIGN_F_NOAUTHENT)) { 
    printf("dnssec_getkey():Key %s not allowed to sign or authenticate %x\n",
	   keyname, pubkey->flags);
    dnssec_freekey(pubkey);
    return NULL;
  }
  if(!(dg_key = (KEY *) malloc(sizeof(KEY)))){
    perror("dnssec_getkey():Memory allocation error");
    exit(42);
  }
  memset(dg_key, 0, sizeof(KEY));
  if( ReadPrivateKey( keyname, dg_key)) {
#ifdef DNS_SECURITY
    dg_key->axfr = do_axfr;
#else
    dg_key->axfr = 0;
#endif
    dg_key->flags = pubkey->flags;
    dg_key->ttl   = 1;  /* I need to set this to something better later XXXX*/

    /*
     * add this key to the chain of existing keys  we add to front as we
     * expect that each key will only be used for short time. 
     * and the chain will be short 
     */
    next_key = (struct key_chain *) malloc( sizeof(struct key_chain)); 
    if( next_key == NULL) {
      perror("dnssec_getkey():Memory allocation error2");
      exit(42);
    }
    next_key->key = dg_key; 
    next_key->next = chain_head; 
    chain_head = next_key;
  }
  else {
    dnssec_freekey(dg_key);
    dg_key = NULL;
  }
  dnssec_freekey(pubkey);
  return dg_key;
}


/*
 * Generate a public/private key pair and store them
 * Parameters 
 *     name   the name of the new key 
 *     bits   the size of the new key in bits 
 *     exp    what exponent to use  0 ==> 3  any other value Fermant4
 *     flags  what DNS KEY flags this key has by default 
 *     protocol the default value of the DNS KEY protocol field
 *     alg    what KEY algorithm to use 
 * Return codes
 *     0 Failure 
 *     1 Success
 */
int
dnssec_genkey(name, bits, exp, flags, protocol, alg)
int bits, exp, flags, protocol, alg;
char *name;
{
  KEY new_key;
  int len; 
  if( name == NULL) 
    return 0; 

  new_key.key_name = strdup(name);
  new_key.key_size = bits;
  new_key.alg      = alg;
  new_key.flags    = flags;
  new_key.KEY_struct= NULL;

  strip_trailing_dots(new_key.key_name);

  if( (flags & SIGN_F_NOKEY) != SIGN_F_NOKEY){ 
    switch( alg) {
    case KEY_RSA:
#ifdef RSAREF
	if(  Generate_RSAKey(&new_key,exp) == 0){
	  printf("dnssec_genkey(): Key generation failure %s %d %d %d\n",
		 new_key.key_name, new_key.alg, new_key.key_size, exp);
	  return 0;
	}
	break;
#endif /* RSAREF */
      default: 
	printf("dnssec_genkey():Unsupported algorithm %d\n",alg);
	return 0;
      }
    dnssec_putkey( &new_key);
  }
  WritePublicKey( &new_key, protocol);
  return 1;
}

/*
 * Function to write out private KEY
 * This function is should only be called by dnssec_genkey() but is included
 * here to allow future extensions such as changing password of keys
 * Parameters: 
 *    name     the name of the File to store the KEY in
 *    private  The Private key to store 
 *    size     Size of the private key
 *    alg      What algorithm this is
 */

static void 
dnssec_putkey( key)
KEY  *key;
{
  u_char encoded_block[RAW_KEY_SIZE],encrypted_block[RAW_KEY_SIZE];
  char file[PATH_MAX];
  int i,len, out_len; 
  /* First I need to encode the key into the portable key format */
  switch ( key->alg){
   case KEY_RSA:
#ifdef RSAREF
      if( (len = Encode_RSAKey( ((RSA_Key *) key->KEY_struct)->Private_Key,
			      encoded_block, sizeof(encoded_block))) <= 0){
	printf( "dnssec_putkey(): Could not encode private RSA key %d\n", len);
	return ;
      }
      break;
#endif /* RSAREF */
  default: 
    printf("dnssec_putkey(): Unknown algorithm %d\n", key->alg);
    return;
  }
  /* Now I can create the file I want to use */
  strip_trailing_dots(key->key_name);
  sprintf(file,"%s.priv",key->key_name);
  if( rename_old( file,
	     "dnssec_putkey():Old Private key file exists Renaming to %s\n"))
      printf( "dnssec_putkey(): Rename Falied %d\n", errno);

  if( (out_len =dnssec_encrypt( encoded_block, encrypted_block, len))>0){
    if( (i=open(file, O_WRONLY| O_CREAT | O_TRUNC, 0600))) {
      int nn;
      if((nn=write(i, encrypted_block, out_len)) != out_len) { 
	printf("dnssec_putkey(): Write failure on %s %d != %d errno=%d\n",
	       file, out_len, nn, errno);
	exit(-3);
      }
      close(i);
    }
    else 
      printf("dnssec_putkey(): Can not create file %s\n",file);
  }
  else 
      printf("dnssec_putkey(): Encryption failed %d %d\n", out_len,len);
  memset(encoded_block, 0, len);
  memset(encrypted_block, 0, len);
}

/* 
 * dnssec_pubkey() Function to convert contents of an KEY RR into a 
 *                 key structure
 * This routine gets passed a name of the key, the algorithm and the beginning
 * of the key data in the KEY RR (byte 4). 
 * It returns a key structure  or NULL if unsuccessfull
 */
KEY * 
dnssec_pubkey(alg, len, key, name)
int alg, len;
u_char *key, *name;
{
  KEY *key_st = (KEY *) malloc( sizeof(KEY));
  if( key_st == NULL) { 
    perror("dnssec_pubkey(): No memory available\n");
    exit(42);
  }
  if( name == NULL) 
    return NULL;
  memset(key_st, 0, sizeof(KEY));
  key_st->alg = alg;
  strip_trailing_dots(name); /* just to be sure */
  switch( alg) { 
    case KEY_RSA:
#ifdef RSAREF  
      if(dns2rsakey(len, key, name, key_st))
	return key_st;
      break; 
#endif /* RSAREF */
  default: 
    printf("dnssec_pubkey(): Unknown alg %d\n",alg);
  }
  SAFE_FREE(key_st);
  return key_st;
}

/* 
 * dnssec_freekey( key) 
 * function to release all data structures pointed to by a key record
 * and the key record passed in 
 */ 
void
dnssec_freekey( f_key)
KEY *f_key;
{
  struct key_chain *curr=NULL, *prev=NULL;
  if(f_key== NULL) 
    return;
  /* if the key is in the chain of keys I should remove it from it */
  for(curr = chain_head; curr && curr->key != f_key; curr = curr->next)
    prev = curr;
  if( curr) { /* key found in chain */ 
    if( prev == NULL) chain_head= curr->next;
    else prev->next = curr->next;
    free(curr);
  }

  switch(f_key->alg) {
    case KEY_RSA: 
#ifdef RSAREF
      Free_RSAKey((RSA_Key *) f_key->KEY_struct);
      f_key->KEY_struct = NULL;
      break; 
#endif /* RSAREF */
    default: 
      printf("dnsfree_key(): Unknown key alg %d\n",f_key->alg);
  }
  if( f_key->KEY_struct) SAFE_FREE((char *)f_key->KEY_struct);
  if( f_key->key_name)   SAFE_FREE(f_key->key_name);
  SAFE_FREE(f_key);
}

/* 
 * dnssec_decrypt function to decrypt the content of the private key file 
 * inputs: ibuf == buffer with encrypted file contents
 *         obuf == buffer to put decrypted file into 
 *         len  == number of bytes in ibuf 
 */ 
static int 
dnssec_decrypt( ibuf, obuf, len)
     u_char *ibuf, *obuf;
     int len;
{
  int tries=3, ret =0, j;
  u_char pass[40];
  char *str="Private-key-format: v";
  while (tries-- && 
	 (ret=get_passwd(pass, "Key decryption Password:", sizeof(pass)))>=0){
    if(ret) des_function( pass, ibuf, obuf, len, DES_DECRYPT);
    else memcpy(obuf,ibuf, len);
    if(!(ret=strncmp(obuf, str, strlen(str)))) {
      memset(pass,0,sizeof(pass));
      j=(int) obuf[len-1];
      return len-j;
    }
  }
  printf("Dnssec_decrypt():Failed decrypting the buffer\n");
  memset(pass,0,sizeof(pass));
  memset(obuf, 0, len);
  return 0;
}
/* 
 * dnssec_encrypt function to encrypt the content of the private key file 
 * inputs: ibuf == buffer with encrypted file contents
 *         obuf == buffer to put encrypted file into 
 *         len  == number of bytes in ibuf 
 */ 
static int 
dnssec_encrypt( ibuf, obuf, len)
     u_char *ibuf, *obuf;
     int len;
{
  u_char pass[2] [40];
  int tries=3, p1,p2;
  memset(pass, 0, sizeof(pass));
  printf("Enter Encryption password for the key, <Return> for no password\n");
  for( tries=3; tries==3 ||(tries >= 0 && strcmp( pass[1], pass[0])); tries--){
    p1 = get_passwd(pass[0], "Key Encryption Password:", sizeof(pass[1]));
    p2 = get_passwd(pass[1], "Retype Key Encryption Password:", 
		    sizeof(pass[1]));
  }
  if( p1 == p2 && !strncmp(pass[1], pass[0], p1)){
    if( p1 > 0 )
      len = des_function( pass[0], ibuf, obuf, len, DES_ENCRYPT);
    else  /* no password */
      memcpy( obuf, ibuf, len);
    memset(pass, 0, sizeof(pass));
  }
  else { 
    printf("dnssec_encrypt(): Non acceptable password\n");
    return 0;
  }
 return len;
}

/* 
 * Function: dnssec_readpubkey() 
 *         standard function to read in a public key from file 
 * Parameters: 
 *     name: name of the key to read in 
 * Return Codes: 
 *     NULL: if the key does not exist or no name is supplied 
 *     KEY *: Initalized key structure if the key exists. 
 * 
 */ 
KEY *
dnssec_readpubkey(name)
char *name;
{ 
  int flags, proto, alg, len, dlen;
  char str[MAXDNAME], enckey[512],deckey[512], *notspace;
  FILE *fp;
  KEY *read_key=NULL;
  
  if( name == NULL) 
    return NULL;
  strip_trailing_dots( name);
  (void) sprintf(str,"%s.public",name);
  if((fp = fopen(str,"r")) == NULL){
    printf("dnssec_readpubkey(): Public Key not found %s\n", str);
    return NULL;
  }
  if(fscanf(fp,"%d %d %d",&flags, &proto, &alg) != 3) { 
    printf("dnssec_readpubkey(): Can not read flag/proto/alg field from %s\n",
	   str);
    return NULL;
  }
  read_key = (KEY *)  malloc(sizeof(KEY));
  if( read_key == NULL) { 
    perror("dnssec_readpubkey():No memory available\n");
    exit( 42);
  }
  memset(read_key, 0, sizeof(KEY));
  read_key->flags = flags;
  read_key->alg   = alg;
  read_key->proto = proto;
  fgets(enckey, sizeof(enckey), fp);
  if( (len = strlen(enckey)) <= 0) { 
    perror("dnssec_readpubkey(): fgets does not get anything ");
    exit(43);
  }
  fclose(fp);
  enckey[--len]= '\0'; /* discard \n */
  for(notspace = enckey;isspace(*notspace); len--, notspace++);

  dlen = decode64( notspace, deckey, len);
  switch (alg) {
  case KEY_RSA: 
#ifdef RSAREF
      if(dns2rsakey(dlen, deckey, name, read_key))
	return read_key;
      break;
#endif
  default: 
    printf("dnssec_readpubkey(): Unknown alg %d\n",alg);
  }
  SAFE_FREE(read_key);
  return NULL;
}


/* 
 * Standard externaly visible routine to write out a public key file 
 * Parameters: 
 *    KEY *: key structure that is not null 
 *    proto: the protocol field in the DNS key structure. 
 */ 
int
dnssec_writepubkey(key, proto)
KEY *key;
int proto;
{
  return WritePublicKey(key, proto);
}
/*
 * Following are functions that actualy deal with low level stuff of keys
 * and know about key formats
 *
 * ReadPrivateKey( name, KEY structure) 
 *      Function reads in private key with from a file named <name>.priv 
 *      Fills out the KEY structure 
 *  return 1 if everthing works 
 *         0 if there is any problem
 *
 */
static int
ReadPrivateKey (name, pk_key)
     char *name;
     KEY *pk_key;
{
  int     fd, cnt, alg, len, major, minor, file_major, file_minor;
  int32_t foot;
  char    filename[200];
  u_char  in_buff[RAW_KEY_SIZE], out_buff[RAW_KEY_SIZE], *p;
  
  /* first check if we can find the key file */ 
  if( strlen(name)<5 || memcmp(&name[strlen(name)-5], ".priv",5))
    sprintf( filename, "%s.priv",name);
  else 
    strcpy( filename, name);
  if ((fd = open(filename, O_RDONLY, 0)) == -1) { 
    printf( "ReadPrivateKey: Could not open file %s in directory %s\n",
	   filename,(char *) getcwd(NULL, PATH_MAX-1));
    return 0;
  }
  /* now read the header info from the file */
  if( (cnt=read(fd, in_buff, sizeof(in_buff))) < 5) {
    close(fd);
    printf("ReadPrivateKey: error reading file %s(empty file)\n", filename);
    return 0;
  }
  close(fd);
/*  printf("Decoding Key %s\n", name); */
  if(memcmp(in_buff, "Private-key-format: v",20)!=0) {  /* File encrypted */
    /*simulate des program */
    if( (len=dnssec_decrypt(in_buff, out_buff, cnt)) <= 0){
      printf("ReadPrivateKey():Failed decrypting key %s\n",name);
      goto fail;
    }
  }
  else 
    memcpy( out_buff, in_buff, len=cnt);
  p=out_buff;
  if (!verify_str((char **) &p, "Private-key-format: v")) {
    printf("ReadPrivateKey(): Not a Key file/Decrypt failed %s\n", name);
    goto fail;
  }
  sscanf(p,"%d.%d",    &file_major, &file_minor);  /* read in file format */
  sscanf(KEY_FILE_FORMAT,"%d.%d",&major, &minor);
  if( file_major < 1 ) {
    printf("ReadPrivateKey(): Unknown keyfile %d.%d version for %s\n",
	   file_major, file_minor, name);
    goto fail;
  }
  else if( file_major > major || file_minor > minor) 
    printf(
      "ReadPrivateKey(): Keyfile %s version higher than mine %d.%d MAY FAIL\n",
	   name, file_major, file_minor);
  while( *p++!='\n');  /* skip to end of line */
  if(!verify_str((char **) &p, "Algorithm: "))
    goto fail;
  if( sscanf(p,"%d", &alg) != 1) 
    goto fail;
  while(*p++ != '\n'); /* skip to end of line */
  pk_key->key_name = (char *) strdup( name);
  strip_trailing_dots(pk_key->key_name);

  switch ( alg) { 
  case KEY_RSA:
#ifdef RSAREF 
    {
      RSA_Key *r_key = (RSA_Key *) malloc( sizeof(RSA_Key ));
      if( r_key ==NULL) {
	perror("ReadPrivateKey():Memory allocation error");
	exit(42);
      }
      memset(r_key, 0, sizeof(RSA_Key));
      foot = Decode_RSAKey(r_key, p, &out_buff[len]-p);
      if( foot < 0 || r_key->Private_Key == NULL) { 
	printf("ReadPrivateKey():Decode_RSAKey failed \n");
	memset(r_key,0, sizeof(RSA_Key));
	SAFE_FREE(r_key);
	goto fail;
      }
      pk_key->key_size = r_key->Private_Key->bits;
      r_key->signer    = pk_key->key_name; 
      pk_key->foot = (u_int16_t) foot;
      pk_key->alg  = alg;
      pk_key->KEY_struct = r_key;
      break;
    }
#endif /* RSAREF */
  default: 
    printf("ReadPrivateKey():Unknown Key type %x\n", alg);
    goto fail;
  }
  memset( in_buff, 0, cnt);
  memset( out_buff,0, cnt); 
  return 1;
fail:
  memset( in_buff,  0, cnt);
  memset( out_buff, 0, cnt);
  return 0;
}

/* 
 * WritePublicKey() function to write to file public key in DNS format
 * Inputs: 
 *    Key : a generic Key structure
 *    proto: The value of the proto field for the public KEY
 * Output:
 *    0 if failed 
 *    1 if successful
 */ 
static int
WritePublicKey( key, proto) 
KEY *key;
int proto;
{
  FILE *fp;
  char filename[PATH_MAX];
  char out_key [RAW_KEY_SIZE];
  int len;

  memset(out_key, 0, sizeof(out_key));
  if(key ==NULL ) {
    printf("WritePublicKey(): No Key specified \n");
    return 0; 
  }
  else if( key->KEY_struct  && 
	  ((key->flags & SIGN_F_NOKEY) != SIGN_F_NOKEY) ) { 
    switch( key->alg) {
    case KEY_RSA: 
#ifdef RSAREF 
      if((len = rsa2dnskey( ((RSA_Key *)key->KEY_struct)->Public_Key, 
			    out_key, sizeof(out_key)))> 0) 
	break;
#endif
    default:
      len = 0;
      printf("Trying to convert an unkown algorithm %d\n", key->alg);
      return;
    }
  }
      /* create public key file */
  strip_trailing_dots(key->key_name); /* better safe than sorry */
  sprintf(filename, "%s.public",key->key_name);
  if( rename_old(filename, "WritePublicKey(): renamed old key %s\n"))
    printf("WritePublicKey():rename_old Failed %s",filename);

  if ((fp = fopen(filename, "w+")) == NULL) {
    printf("Write_PublicKeys:open of file:%s failed (errno=%d)\n", 
	   filename, errno);
    exit(-12);
  }
           /*write out key */
  fprintf(fp,"%d %d %d %s\n",key->flags, proto, key->alg,out_key);
  fclose(fp);
  return 1;
}

/* 
 * dnssec_encpubkey() Function to encode public key into DNS format in
 *   base64 encoding.
 * Input: 
 *     key *: Key structure to encode 
 *     out_storage: array to write the key to 
 *     out_len : size of the output array 
 * Output: 
 *     0: failure 
 *    >0: number of bytes written to out_storage. 
 */
int 
dnssec_encpubkey( key, out_storage, out_len) 
KEY *key;
char *out_storage;
int out_len;
{
  if( key == NULL) 
    return 0;
  switch (key->alg){
    case KEY_RSA:
#ifdef RSAREF
    {
      RSA_Key *rkey = (RSA_Key *) key->KEY_struct;
      return rsa2dnskey(rkey->Public_Key, out_storage, out_len);
    }
#endif 
     default: 
       printf("dnssec_enpubkey(): Unsupported ALG %d\n",key->alg);
  }
  return 0;
}

/* 
 * function to strip any trailing dots from a name
 * Input: 
 *     keyname: character string that is writeable
 * 
 */
static void 
strip_trailing_dots( char *keyname) 
{ 
  int t;
  if( keyname) {
    t = strlen(keyname);  /* make sure no keyname ends with a . */
    while( t && keyname[--t] == '.') 
      keyname[t] = '\0'; 
  }
  return; 
}
